#property copyright "ArteInVolo"
#property version "1.00"

#property indicator_separate_window 
#property indicator_buffers 1
#property indicator_plots 1

#property indicator_type1 DRAW_LINE
#property indicator_color1 clrBlue
#property indicator_style1 STYLE_SOLID
#property indicator_width1 1
#property indicator_label1 "Index"

// *** Input *** //

input int InpDepth = 200;
input int InpIterations = 100;
input double InpPrecision = 0.001;
input string InpSymbols = "AUDUSD,EURUSD,GBPUSD,USDCHF,USDJPY,USDCAD,NZDUSD,SILVER,USDSEK,USDSGD,USDMXN,USDNOK";

// *** Globals *** //

struct SSeries
{
    string mName;
    double mVar;
    double mMean;
    double mCovs[];
    double mCors[];
    double mPoints[];
    double mVectors[];
    double mSyncCloses[];
    datetime mSyncTimes[];
};

long iChartId;
double iChart1[];
int iSubWindowId;
SSeries iSeries[];

// *** Implementation *** //

int OnInit()
{
    SetIndexBuffer(0, iChart1, INDICATOR_DATA);
    ArrayInitialize(iChart1, 0);
    ArraySetAsSeries(iChart1, true);

    IndicatorSetInteger(INDICATOR_DIGITS, _Digits);
    IndicatorSetString(INDICATOR_SHORTNAME, "Charts");
    
    ListSymbols(iSeries);
    Setup(iSeries, ArraySize(iSeries), InpDepth, true);
    
    iChartId = ChartID();
    iSubWindowId = ChartWindowFind(iChartId, "Charts");
    
    EventSetTimer(1);
    
    return(INIT_SUCCEEDED);
}

void OnTimer()
{
    Display(iSeries);
}

int OnCalculate(
    const int bars,
    const int counted,
    const datetime &time[],
    const double &open[],
    const double& high[],
    const double& low[],
    const double &close[],
    const long &tickVolume[],
    const long &volume[],
    const int &spread[])
{
    return(bars);
}

void Setup(SSeries& series[], const int rows, const int cols, const bool clear)
{
    ArrayResize(series, rows);
    
    for (int k = 0; k < rows; k++)
    {
        ArrayResize(series[k].mCovs, cols);
        ArrayResize(series[k].mCors, cols);
        ArrayResize(series[k].mPoints, cols);
        ArrayResize(series[k].mVectors, cols);
        ArrayResize(series[k].mSyncTimes, cols);
        ArrayResize(series[k].mSyncCloses, cols);
        
        if (clear)
        {
            ArrayFill(series[k].mCovs, 0, cols, 0);
            ArrayFill(series[k].mCors, 0, cols, 0);
            ArrayFill(series[k].mPoints, 0, cols, 0);
            ArrayFill(series[k].mVectors, 0, cols, 0);
            ArrayFill(series[k].mSyncTimes, 0, cols, 0);
            ArrayFill(series[k].mSyncCloses, 0, cols, 0);
        }
    }
}

void Display(SSeries& series[])
{
    int index = 0;
    int depth = InpDepth;
    int order = ArraySize(series);

    if (Synchronize(series, depth))
    {
        SSeries eigens[], vectors[];
        Setup(eigens, order, order, true);
        Setup(vectors, order, order, true);
        GetBaseMatrix(series, depth);
        GetCovMatrix(series, depth);
        GetEigenValues(series, vectors, eigens);
        ShowVector(eigens, vectors, true);
        
        for (int k = depth - 1; k > 0; k--)
        {
            iChart1[index] = 1;
            index++;
        }
    }
}

double GetBarByTime(const string symbol, const datetime time)
{
    datetime last;
    SeriesInfoInteger(symbol, _Period, SERIES_LASTBAR_DATE, last);
    return Bars(symbol, _Period, time, last) - 1;
}

int Synchronize(SSeries& series[], const int depth)
{
    int order = ArraySize(series);

    for (int k = 0; k < order; k++)
    {
        double closes[];
        int copied = CopyClose(series[k].mName, _Period, 0, depth, closes);
        
        if (copied > 0)
        {
            datetime currentTime = GetLastTime(series);
            int closeIndex = ArraySize(closes) - 1;
            series[k].mSyncCloses[depth - 1] = closes[closeIndex];
        
            for (int n = depth - 1; n > 0; n--)
            {
                if (GetBarByTime(series[k].mName, currentTime) > 0 && closeIndex > 0)
                {
                    closeIndex--;
                }

                series[k].mSyncCloses[n] = closes[closeIndex];
                series[k].mSyncTimes[n] = currentTime;
                currentTime -= PeriodSeconds();
            }
            
            series[k].mSyncCloses[0] = series[k].mSyncCloses[closeIndex];
        }
    
        if (copied < 1) 
        {
            return(0);
        }
    }
    
    return(1);
}

void ListSymbols(SSeries& series[])
{
    string symbols[];
    int index = 0;
    int count = StringSplit(InpSymbols, ',', symbols);

    for (int k = 0; k < count; k++)
    {
        if (SymbolSelect(symbols[k], true))
        {
            ArrayResize(series, index + 1); 
            series[index++].mName = symbols[k];
        }
    }
}

datetime GetLastTime(SSeries& series[])
{
    datetime startTime = 0;
    int count = ArraySize(series);

    for (int k = 0; k < count; k++)
    {
        datetime times[];
        CopyTime(series[k].mName, _Period, 0, 1, times);
        startTime = startTime > times[0] ? startTime : times[0];
    }

    return(startTime);
}

void GetBaseMatrix(SSeries& series[], const int depth)
{
    int order = ArraySize(series);

    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < depth; n++)
        {
            series[k].mPoints[n] = 1000 * MathLog(series[k].mSyncCloses[depth - n - 1]);
        }
    }
}

void GetMeans(SSeries& series[], const int depth)
{
    int order = ArraySize(series);
    
    for (int k = 0; k < order; k++)
    {
        series[k].mMean = 0;
        
        for (int n = 0; n < depth; n++)
        {
            series[k].mMean += series[k].mPoints[n];
        }
        
        series[k].mMean = series[k].mMean / depth;
        
        for (int n = 0; n < depth; n++)
        {
            series[k].mPoints[n] = series[k].mPoints[n] - series[k].mMean;
        }
    }
}

void GetCovMatrix(SSeries& series[], const int depth)
{
    double variance = 0;
    int order = ArraySize(series);
    
    GetMeans(series, depth);
    
    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            series[k].mCovs[n] = 0;

            for (int i = 0; i < depth; i++)
            {
                series[k].mCovs[n] += series[k].mPoints[i] * series[n].mPoints[i];
            }
            
            series[k].mCovs[n] /= depth;
            
            if (k == n)
            {
                series[k].mVar = series[k].mCovs[n];
            }
        }
    }
}

double GetMaxInLuMatrix(SSeries& series[], int& X, int& Y)
{
    double max = 0;
    int order = ArraySize(series);
    
    for (int k = 0; k < order; k++)
    {
        for (int n = k + 1; n < order; n++)
        {
            double candidate = MathAbs(series[k].mPoints[n]);
            
            if (candidate > max)
            {
                max = candidate;
                X = k;
                Y = n;
            }
        }
    }
    
    return max;
}

void MultiplyMatrix(SSeries& series[], SSeries& A[], SSeries& B[], const bool transposeFirst)
{
    int order = ArraySize(A);
    
    SSeries results[];
    Setup(results, order, order, true);

    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            results[k].mPoints[n] = 0;

            for (int i = 0; i < order; i++) 
            {
                results[k].mPoints[n] += transposeFirst ? A[i].mPoints[k] * B[i].mPoints[n] : A[k].mPoints[i] * B[i].mPoints[n];
            }
        }
    }
    
    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            series[k].mPoints[n] = results[k].mPoints[n];
        }
    }
}

double RotateMatrix(SSeries& series[], SSeries& rotations[], int X, int Y)
{
    int order = ArraySize(series);
    double range = series[X].mPoints[X] - series[Y].mPoints[Y];

    SSeries multipliers[];
    Setup(multipliers, order, order, true);
    
    if (range == 0)
    {
        return(0);
    }

    double angle = MathArctan(2 * series[X].mPoints[Y] / range) / 2;
    double angleSin = MathSin(angle);
    double angleCos = MathCos(angle);

    for (int k = 0; k < order; k++) 
    {
        ArrayFill(rotations[k].mPoints, 0, order, 0);
        ArrayFill(multipliers[k].mPoints, 0, order, 0);
        rotations[k].mPoints[k] = multipliers[k].mPoints[k] = (k == X || k == Y) ? angleCos : 1;
    }
    
    rotations[Y].mPoints[X] = multipliers[Y].mPoints[X] =  angleSin;
    rotations[X].mPoints[Y] = multipliers[X].mPoints[Y] = -angleSin;
    
    MultiplyMatrix(series, multipliers, series, true);
    MultiplyMatrix(series, series, multipliers, false);
    
    return(1);
}

void GetEigenValues(SSeries& series[], SSeries& vectors[], SSeries& eigens[])
{
    int order = ArraySize(series), row = 0, col = 0;

    SSeries rotations[];
    Setup(rotations, order, order, true);
    
    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            eigens[k].mPoints[n] = series[k].mCovs[n];
        }
        
        eigens[k].mName = series[k].mName;
    }

    for (int k = 0; k < InpIterations; k++)
    {
        double max = GetMaxInLuMatrix(eigens, row, col);
        
        if (max < InpPrecision) 
        {
            break;
        }
        
        if (k > 0)
        {
            RotateMatrix(eigens, rotations, row, col);
            MultiplyMatrix(vectors, vectors, rotations, false);
        }
        else
        {
            RotateMatrix(eigens, vectors, row, col);
        }
    }
}

void DebugMatrix(SSeries& series[])
{
    string result = "\n";
    int order = ArraySize(iSeries);
    
    for (int i = 0; i < order; i++)
    {
       for (int j = 0; j < order; j++)
       {
           result += DoubleToString(series[i].mPoints[j], 5) + " ";
       }
       
       result += "\n";
    }
    
    Comment(result);
}

void ShowVector(SSeries& eigens[], SSeries& vectors[], bool minimize)
{
    double issue = minimize ? 1000000 : 0;
    int index = 0, order = ArraySize(eigens);

    for (int k = 0; k < order; k++) 
    {
        double eigen = MathAbs(eigens[k].mPoints[k]);
    
        if (eigen < issue)
        {
            issue = eigen;
            index = k;
        }
    }

    double max = 0;    
    datetime times[];
    int X = 10, Y = 5, size = 8, precision = 5, offset = int(ChartGetDouble(0, CHART_SHIFT_SIZE) * ChartGetInteger(0, CHART_WIDTH_IN_PIXELS) / 100);

    for (int k = 0; k < order; k++) 
    {
        max = MathAbs(vectors[k].mPoints[index]) > max ? MathAbs(vectors[k].mPoints[index]) : max;
    }

    int maxPixel = offset - 155;

    for (int k = 0; k < order; k++) 
    {
        string alignment = (vectors[k].mPoints[index] > 0 ? " " : "") + DoubleToString(vectors[k].mPoints[index], precision);
        string label = StringFormat("%s  | %s", eigens[k].mName, alignment);
        int level = int(MathAbs(vectors[k].mPoints[index]) * maxPixel / max);

        ShowItem(OBJ_LABEL, eigens[k].mName + "A", label, X, Y, size, clrBlack);
        ShowItem(OBJ_RECTANGLE_LABEL, eigens[k].mName + "B", eigens[k].mName, X + level + 105, Y + 2, size, vectors[k].mPoints[index] > 0 ? clrDeepSkyBlue : clrTomato, level, 10);
        
        Y += size * 2;
    }
}

void ShowItem(const ENUM_OBJECT type, const string name, const string value, const int X, const int Y, const int size, const color clr, const int W = 0, const int H = 0)
{
    if (ObjectFind(iChartId, name) < 0)
    {
        ObjectCreate(iChartId, name, type, iSubWindowId, 0, 0);
    }

    ObjectSetInteger(iChartId, name, OBJPROP_ANCHOR, ANCHOR_RIGHT_UPPER);
    ObjectSetInteger(iChartId, name, OBJPROP_CORNER,CORNER_RIGHT_UPPER);
    ObjectSetInteger(iChartId, name, OBJPROP_FONTSIZE, size);
    ObjectSetInteger(iChartId, name, OBJPROP_XDISTANCE, X);
    ObjectSetInteger(iChartId, name, OBJPROP_YDISTANCE, Y);
    ObjectSetInteger(iChartId, name, OBJPROP_COLOR, clr);
    ObjectSetString(iChartId, name, OBJPROP_TEXT, value);
    
    if (type == OBJ_RECTANGLE_LABEL) 
    {
        ObjectSetInteger(iChartId, name, OBJPROP_BGCOLOR, clr);
        ObjectSetInteger(iChartId, name, OBJPROP_BACK, false);
        ObjectSetInteger(iChartId, name, OBJPROP_FILL, true);
        ObjectSetInteger(iChartId, name, OBJPROP_WIDTH, 0);
        ObjectSetInteger(iChartId, name, OBJPROP_YSIZE, H);
        ObjectSetInteger(iChartId, name, OBJPROP_XSIZE, W);
    }
}